#include "inc/ffs.h"
#include "close.tmh"
NTSTATUS
Close (
	PDEVICE_OBJECT		PtrDeviceObject,
	PIRP					PtrIrp)

{
	NTSTATUS   Result = STATUS_INVALID_PARAMETER;
	PIO_STACK_LOCATION PtrIrpSp = NULL;
	PFILE_OBJECT			 PtrOldFO = NULL;
	PTR_T_VCB          PtrVCB;
  BOOLEAN            ResourceAcq = FALSE;
  PTR_T_FCB          PtrFCB;
  PTR_T_CCB          PtrCCB;
  KIRQL               Irql;
  PTR_T_VOLUME       PtrVolume;
	__try
	{
		DoTraceMessage(FFS_CREATE_CLOSE,"Close request");
		if (PtrDeviceObject == GlobalData.PtrMainDeviceObject)
		{
				DoTraceMessage(FFS_CREATE_CLOSE,"Close main device");
			Result = STATUS_SUCCESS;
			__leave;
		}
		PtrIrpSp     = IoGetCurrentIrpStackLocation(PtrIrp);
		PtrOldFO     = PtrIrpSp->FileObject;
		PtrVolume = PtrDeviceObject->DeviceExtension;
		if (PtrVolume->Magic!=VOLUME_MAGIC) __leave;
		ExAcquireResourceExclusiveLite(&(PtrVolume->MainResource),TRUE);
		ResourceAcq = TRUE;
		Result = CheckBoth(PtrOldFO->FsContext2);
		if (!NT_SUCCESS(Result)) __leave;
		PtrCCB = (PTR_T_CCB)(PtrOldFO->FsContext2);
		PtrFCB = PtrCCB->PtrParentFCB;
		PtrVCB = PtrFCB->PtrParentVCB;

		Result = DeleteCCB(PtrCCB);
		if (!NT_SUCCESS(Result)) __leave;
		DoTraceMessage(FFS_CREATE_CLOSE,"Close for inode %u ok,\
		 reference  count is %u ",PtrFCB->InodeNumber,PtrFCB->ReferenceCount);
		if ((PtrFCB->ReferenceCount==0) &&  (PtrFCB->OpenHandleCount==0)
		&& (PtrFCB != &(PtrVolume->VolumeFCB)) )
		{
			DoTraceMessage(FFS_CREATE_CLOSE,"Deleting FCB");
			Result =DeleteFCB(PtrFCB);
		}
		if (!NT_SUCCESS(Result)) __leave;
		DoTraceMessage(FFS_CREATE_CLOSE,"File open count %u",PtrVolume->FileOpenCount);
		if 	((PtrVolume->Dismount) && (PtrVolume->FileOpenCount<1))
		{
			DoTraceMessage(FFS_CREATE_CLOSE,"Dismouting volume");
			IoAcquireVpbSpinLock(&Irql);
			CLEAR_FLAG(PtrVolume->PtrVPB->Flags,VPB_MOUNTED);
			IoReleaseVpbSpinLock(Irql);
			Result  = STATUS_SUCCESS;
			ExReleaseResourceLite(&(PtrVolume->MainResource));
		  ResourceAcq = FALSE;
			DeleteVolume(PtrVolume);
			IoDeleteDevice(PtrDeviceObject);
		}

		Result = STATUS_SUCCESS;
	}
	__finally
	{
		if (ResourceAcq) ExReleaseResourceLite(&(PtrVolume->MainResource));
	}
	PtrIrp->IoStatus.Status=Result;
	IoCompleteRequest(PtrIrp, IO_NO_INCREMENT);
	return Result;
}