#include "inc/ffs.h"
#include "fscontrol.tmh"
/*****************************************************************************
Zabezpeuje zamknutie jednotky
*****************************************************************************/
NTSTATUS
LockVolume (
	PDEVICE_OBJECT		PtrDeviceObject,
	PIRP					PtrIrp)
{
	NTSTATUS            Result = STATUS_DRIVER_INTERNAL_ERROR;
	PIO_STACK_LOCATION  PtrIrpSp;

	DoTraceMessage(FFS_FSCONTROL,"Lock volume entry point");
	Result = STATUS_SUCCESS;
	PtrIrpSp = IoGetCurrentIrpStackLocation(PtrIrp);
	PtrIrp->IoStatus.Status = Result;
	IoCompleteRequest(PtrIrp, IO_NO_INCREMENT);
	return Result;
}/*LockVolume*/
/*****************************************************************************
*Zabezpeuje odmontovanie jednoky
*****************************************************************************/
NTSTATUS
DismountVolume (
	PDEVICE_OBJECT		PtrDeviceObject,
	PIRP			PtrIrp)
{
	PIO_STACK_LOCATION  PtrIrpSp;
	KIRQL               Irql;
	PTR_T_VOLUME        PtrVolume;
	NTSTATUS            Result = STATUS_DRIVER_INTERNAL_ERROR;


	DoTraceMessage(FFS_FSCONTROL,"Dismount volume entry point");
	PtrIrpSp = IoGetCurrentIrpStackLocation(PtrIrp);
	PtrVolume = PtrDeviceObject->DeviceExtension;
// akosi nezamyka hehe
	IoAcquireVpbSpinLock(&Irql);
	CLEAR_FLAG(PtrVolume->PtrVPB->Flags,VPB_MOUNTED);
	IoReleaseVpbSpinLock(Irql);
	/*Result  = STATUS_SUCCESS;
	Result = DeleteVCB(PtrVCB);
	IoDeleteDevice(PtrDeviceObject);*/
	Result  = STATUS_SUCCESS;
	PtrVolume->Dismount = TRUE;
	PtrIrp->IoStatus.Status = Result;
	IoCompleteRequest(PtrIrp, IO_NO_INCREMENT);
	return Result;
}/*Dismount volume*/
/*************************************************************************
* Tto funkcia spracva daie poiadavky z fscontrol
*************************************************************************/
NTSTATUS UserFsRequest(
		PDEVICE_OBJECT		PtrDeviceObject,
		PIRP					PtrIrp)

{
	NTSTATUS            Result = STATUS_DRIVER_INTERNAL_ERROR;
	ULONG               FsControlCode;
	PIO_STACK_LOCATION  PtrIrpSp;

	//trebalo by pouvaova kde presne by toto malo by
	PtrIrpSp = IoGetCurrentIrpStackLocation(PtrIrp);
	FsControlCode=PtrIrpSp->Parameters.DeviceIoControl.IoControlCode;
	//FileSystemControl truktra nie je definovab ete e tto vyzer
	//tak isto !!!!!!!!!!!!!!!
	//S to .......
	//zobrat z irpsp2
	switch (FsControlCode)
	{
		case FSCTL_DISMOUNT_VOLUME:
		Result = DismountVolume(PtrDeviceObject, PtrIrp);
		break;

		case FSCTL_LOCK_VOLUME:
		Result = LockVolume(PtrDeviceObject,PtrIrp);
		break;

		case FSCTL_IS_VOLUME_MOUNTED:
		Result =VerifyVolume(PtrDeviceObject,PtrIrp);
		break;


		default:
		PtrIrp->IoStatus.Status = STATUS_DRIVER_INTERNAL_ERROR;
		IoCompleteRequest(PtrIrp, IO_NO_INCREMENT);
		Result = STATUS_DRIVER_INTERNAL_ERROR;
	}
	return Result;
}/*UserFsRequest*/
/*************************************************************************
* Tto funkcia primontuje dan partciu
* Ze je irp a DEviceObject spravne garantuje volajuci
*************************************************************************/
NTSTATUS
MountVolume (
	PDEVICE_OBJECT		PtrDeviceObject,
	PIRP			PtrIrp)

{
  PIO_STACK_LOCATION  PtrIrpSp;
  NTSTATUS            Result = STATUS_UNRECOGNIZED_VOLUME;
  int                 Step = 0;
  PDEVICE_OBJECT      PtrVolumeDevice;       /* Zariadenie vytvor n ovlda */
  PDEVICE_OBJECT      PtrTargetDevice;  /* Partcia na ktor sa pe        */
  PVPB                PtrVPB = NULL;  /* truktra io manara o partcii   */
  ERESOURCE_THREAD    Thread;
  T_VCB              *PtrVCB;

  try
  {
    DoTraceMessage(FFS_FSCONTROL,"Mount volume entry point");
    DoTraceMessage(FFS_FSCONTROL,"Attemp to mount slice number");
    if (PtrDeviceObject != GlobalData.PtrMainDeviceObject)
    {
      Result = STATUS_INVALID_DEVICE_REQUEST;
      DoTraceMessage(FFS_FSCONTROL,"Error cannot mout to main device");
      __leave;
    }
    Step++;/* 1 */
     // akceptuj iba poiadavky pre hlavn zariadenie

    PtrIrpSp = IoGetCurrentIrpStackLocation(PtrIrp);
    PtrTargetDevice = PtrIrpSp->Parameters.MountVolume.DeviceObject;
    PtrVPB = PtrIrpSp->Parameters.MountVolume.Vpb;


    Result = JeFfs(PtrTargetDevice);
    if (!NT_SUCCESS(Result))
    {
      DoTraceMessage(FFS_FSCONTROL,"Ffs not detected");
      __leave;
    }
    DoTraceMessage(FFS_FSCONTROL,"FFs detected");
    Step++;/* 2 */
    //Otestovnanie jednotky na prtomnos ffs
    ExAcquireResourceExclusiveLite(&(GlobalData.MainResource), TRUE);
    Result = IoCreateDevice(
            GlobalData.PtrMainDriverObject,
            sizeof(T_VOLUME),
            NULL,
            FILE_DEVICE_DISK_FILE_SYSTEM,
            0,
            FALSE,
            &PtrVolumeDevice
            );
    if (!NT_SUCCESS(Result))
    {
       DoTraceMessage(FFS_FSCONTROL,"Failed to create slice device , result %u",Result);
       __leave;
    }
    DoTraceMessage(FFS_FSCONTROL,"Created slice device");
    Step++;/* 3 */
    PtrVolumeDevice->StackSize = (CCHAR)(PtrTargetDevice->StackSize + 1);
    Result=InitVolume(PtrVolumeDevice,PtrTargetDevice,PtrVPB);
    if (!NT_SUCCESS(Result))
    {
      DoTraceMessage(FFS_FSCONTROL,"Volume structure initialization failed, result %u",Result);
      __leave;
    }
    DoTraceMessage(FFS_FSCONTROL,"Volume structure initialization ok");
    Step++;/* 4 */
    Step++;/* 5 */
    (PtrIrpSp->Parameters.MountVolume.Vpb)->DeviceObject = PtrVolumeDevice;
    //Spojil som VPB s mojim objektom

    SET_FLAG(PtrVolumeDevice->Flags,DO_BUFFERED_IO);
    CLEAR_FLAG(PtrVolumeDevice->Flags, DO_DEVICE_INITIALIZING);
    DoTraceMessage(FFS_FSCONTROL,"Mount ok");
    }
   __finally
   {
     if ((!NT_SUCCESS(Result)) || (AbnormalTermination()))
     {
       // Switch pokrauje - ke bude krok 3 vykon sa aj 2.
       DoTraceMessage(FFS_FSCONTROL,"Mount interrupted on step %u",Step);
       switch (Step)
       {
          case 5:

          case 3:
	    IoDeleteDevice(PtrVolumeDevice);

       } /*switch*/
      }	/*if*/
      if (Step>1)
         ExReleaseResourceLite( &(GlobalData.MainResource));
   }/*finally*/
   PtrIrp->IoStatus.Status = Result;
   IoCompleteRequest(PtrIrp, IO_NO_INCREMENT);
   return Result;
}
/*********************************************************************************
Ci je jednotka primontovana
**********************************************************************************/
NTSTATUS  VerifyVolume(
		PDEVICE_OBJECT		PtrDeviceObject,
		PIRP					PtrIrp)
{
	PTR_T_VOLUME PtrVolume;
	int       Step = 0;
  NTSTATUS  Result =STATUS_INVALID_DEVICE_REQUEST ;
  UINT32    PartNumber =0;
  UINT32    Move = 0;
  KIRQL     Irql;
  UINT32    Position;
  T_DISK_LABEL DiskLabel;
	__try
	{
		PtrVolume = PtrDeviceObject->DeviceExtension;
		if (PtrDeviceObject == GlobalData.PtrMainDeviceObject) __leave;
		if (PtrVolume->Magic!=VOLUME_MAGIC) __leave;
		DoTraceMessage(FFS_FSCONTROL,"Good volume for verify volume");
		ExAcquireResourceExclusiveLite(&(PtrVolume->MainResource),TRUE);
		Step++;/*1*/

		/*if (!FLAG_ON(PtrVolume->PtrTargetDevice->Flags, DO_VERIFY_VOLUME))
		{
			Result = STATUS_SUCCESS;
		   __leave;
		}*/
		Result = STATUS_SUCCESS;
		if (PtrVolume->Dismount) __leave;

		Result = ReadPhysical(PtrVolume->PtrTargetDevice,
													LABELSECTOR*SECTOR_SIZE+LABELOFFSET,
													sizeof(DiskLabel),
			                    &DiskLabel);

    if (sizeof(DiskLabel)==RtlCompareMemory
    (&(PtrVolume->DiskLabel),&DiskLabel,sizeof(DiskLabel)))

		{
			//!!!!
			CLEAR_FLAG(PtrVolume->PtrTargetDevice->Flags, DO_VERIFY_VOLUME);
		  Result = STATUS_SUCCESS;

		  __leave;
		}
		else
		{
			PtrVolume->Dismount = TRUE;
			IoAcquireVpbSpinLock(&Irql);
			CLEAR_FLAG(PtrVolume->PtrVPB->Flags,VPB_MOUNTED);
			IoReleaseVpbSpinLock(Irql);
			/*ExReleaseResourceLite(&(PtrVolume->MainResource));
			Step--;
			DeleteVCB(PtrVCB);
			IoDeleteDevice(PtrDeviceObject);*/
			Result =STATUS_WRONG_VOLUME;
		}
	}
	__finally
	{
		if (Step>0) ExReleaseResourceLite(&(PtrVolume->MainResource));
	}
	PtrIrp->IoStatus.Status = Result;
	IoCompleteRequest(PtrIrp, IO_NO_INCREMENT);
	return Result;

}/*VerifyVolume*/
/*********************************************************************************
Keby od ns chcel niekto nieo primontova
**********************************************************************************/
NTSTATUS RiadenieFs(
		PDEVICE_OBJECT		PtrDeviceObject,
		PIRP					PtrIrp)

{
	NTSTATUS  Result = STATUS_UNRECOGNIZED_VOLUME;

	//trebalo by pouvaova kde presne by toto malo by
	DoTraceMessage(FFS_FSCONTROL,"FS control entry point");
	switch (IoGetCurrentIrpStackLocation(PtrIrp)->MinorFunction)
	{
		case IRP_MN_MOUNT_VOLUME:
		Result = MountVolume(PtrDeviceObject, PtrIrp);
		break;

		case IRP_MN_USER_FS_REQUEST:
		Result = UserFsRequest(PtrDeviceObject,PtrIrp);
		break;

		case IRP_MN_VERIFY_VOLUME:
		Result =VerifyVolume(PtrDeviceObject,PtrIrp);
		break;

		default:
		PtrIrp->IoStatus.Status = STATUS_DRIVER_INTERNAL_ERROR;
		IoCompleteRequest(PtrIrp, IO_NO_INCREMENT);
		Result = STATUS_DRIVER_INTERNAL_ERROR;
	}
	return Result;
}/* RiadenieFs */
/*********************************************************************
**********************************************************************/
NTSTATUS GetPart(IN PDEVICE_OBJECT ObjektZariadeniaParticie,
                 OUT UINT64  *PtrStartByte)
{
	char Buffer[2048];
	PIRP            PtrIrp;
	IO_STATUS_BLOCK IoStatus;
	KEVENT          Event;
	NTSTATUS        Result;
	PARTITION_INFORMATION *PtrPartInfo;

	if (FLAG_ON(ObjektZariadeniaParticie->Characteristics,FILE_FLOPPY_DISKETTE))
	{
           *PtrStartByte = 0;
	   return STATUS_SUCCESS;
	}
	KeInitializeEvent(&Event, NotificationEvent, FALSE);
	PtrIrp = IoBuildDeviceIoControlRequest(
		  IOCTL_DISK_GET_PARTITION_INFO,
		  ObjektZariadeniaParticie,
		  NULL,
		  0,
		  Buffer,
		  2048,
		  FALSE,
		  &Event,
		  &IoStatus);
	if (!PtrIrp) return STATUS_INSUFFICIENT_RESOURCES;
	Result = IoCallDriver(ObjektZariadeniaParticie, PtrIrp);
	if (!NT_SUCCESS(Result)) return Result;
	Result = KeWaitForSingleObject(&Event, Executive, KernelMode, FALSE, NULL);
	if (!NT_SUCCESS(Result)) return Result;
	PtrPartInfo = (PARTITION_INFORMATION *)Buffer;
	*PtrStartByte = PtrPartInfo->StartingOffset.QuadPart;
	return Result;
}

/*************************************************************************
Procedra o overuje i je partcia skutone bsd
**************************************************************************/
NTSTATUS JeFfs ( PDEVICE_OBJECT ObjektZariadeniaParticie)
{
	T_DISK_LABEL     Menovka;
	T_SUPER_BLOCK    SubSystem;
	NTSTATUS  Result=STATUS_UNRECOGNIZED_VOLUME;
  int              i;
  int              j;
  UINT64           StartByte;
  UINT64           SearchList[5];
  UINT64           Position;

  DoTraceMessage(FFS_FSCONTROL,"Testing ffs presence on slice number");
  SearchList[0] = SBLOCK_UFS2;
  SearchList[1] = SBLOCK_UFS1;
  SearchList[2] = SBLOCK_FLOPPY;
  SearchList[3] = SBLOCK_PIGGY;
  SearchList[4] = -1;

  DoTraceMessage(FFS_FSCONTROL,"Reading disklabel from byte %u",
                 SECTOR_SIZE*LABELSECTOR+LABELOFFSET);
  Result = ReadPhysical(ObjektZariadeniaParticie,
    LABELSECTOR*SECTOR_SIZE+LABELOFFSET,
    sizeof(Menovka),
    &Menovka);
  if ((Menovka.d_magic!=DISKMAGIC) ||( !NT_SUCCESS(Result)) )
  {
     DoTraceMessage(FFS_FSCONTROL,"Disklabel magic number invalid,or  result %u bad",Result);
     return Result;
  }
  DoTraceMessage(FFS_FSCONTROL,"Valid disklabel detected");
  Result = GetPart(ObjektZariadeniaParticie,&StartByte);
  if (!NT_SUCCESS(Result))
  {
    DoTraceMessage(FFS_FSCONTROL,"Slice info failed, result %u",Result);
    return Result;
  }
  DoTraceMessage(FFS_FSCONTROL,"Slice info ok, absolute  slice start is %I64u byte",StartByte);
  for (j=0;j<MAXPARTITIONS;j++)
  if (Menovka.d_partitions[j].p_fstype==FS_BSDFFS)
  {
    for (i=0;i<5;i++)
    {
      DoTraceMessage(FFS_FSCONTROL,"Partition %u has BSD_FFS flag",j);
       DoTraceMessage(FFS_FSCONTROL,"Superblock loc. from disklabel1 is %I64u in sectors",
       Menovka.d_partitions[j].p_offset);
      Position = (Menovka.d_partitions[j].p_offset);
      Position = Position*SECTOR_SIZE;
      DoTraceMessage(FFS_FSCONTROL,"Superblock loc. from disklabel is %I64u",Position);
      Position = Position +SearchList[i];
      DoTraceMessage(FFS_FSCONTROL,"Superblock loc. from disklabel is %I64u",Position);
      Position = Position -StartByte;
      DoTraceMessage(FFS_FSCONTROL,"Reading superblock from byte %I64u",Position);
      Result = ReadPhysical(ObjektZariadeniaParticie,
                            Position,
			    sizeof(SubSystem),
                            &(SubSystem));
      if (!NT_SUCCESS(Result))
      {
        DoTraceMessage(FFS_FSCONTROL,"Superblock read failed, result %u",Result);
        return Result;
      }
      if ((SubSystem.fs_magic==FS_UFS1_MAGIC) ||
         (SubSystem.fs_magic==FS_UFS2_MAGIC ))
      {
	 Result = STATUS_SUCCESS;
         DoTraceMessage(FFS_FSCONTROL,"Superblock magic ok");
	 return Result;
      }
      else
      {
         DoTraceMessage(FFS_FSCONTROL,"FFS flag present, but superblok magic invalid on part %d",i);
         Result = STATUS_UNRECOGNIZED_VOLUME;
      }
    }/*for i*/
  }/*for j*/
  return (Result);
}/*JeFfs*/
