#include "inc/ffs.h"
#include "mismas.tmh"

/*********************************************************************************
*
* Sem prde kad irp, ktor ma tento ovlda spracova.
* Tie sa tu er vnimky.
*
*********************************************************************************/

NTSTATUS Rozhod(
		PDEVICE_OBJECT	PtrDeviceObject,   // ono to me prs na hocijak
		PIRP					PtrIrp)		  // ktor tento ovlda obhospodaruje
{
	NTSTATUS       Result = STATUS_INVALID_PARAMETER;
	KIRQL         Ehm;
	PTR_T_VOLUME  PtrVolume = NULL;


	Ehm =  KeGetCurrentIrql();
	if (Ehm != PASSIVE_LEVEL)
	{

           DoTraceMessage(FFS_CHECKS,"To high irq level");
           //pre moj pokoj zatial iba takto
           PtrIrp->IoStatus.Status = STATUS_DRIVER_INTERNAL_ERROR;
	   IoCompleteRequest(PtrIrp, IO_NO_INCREMENT);
	   return STATUS_DRIVER_INTERNAL_ERROR;
	}
	FsRtlEnterFileSystem();

	__try
	{
		Result = CheckDevIrp (PtrDeviceObject,PtrIrp);
                if (!NT_SUCCESS(Result))
		{
		  DoTraceMessage(FFS_CHECKS,"Aborting request");
		  __leave;
		}
		switch (IoGetCurrentIrpStackLocation(PtrIrp)->MajorFunction)
		{
			case IRP_MJ_CREATE:
		  Result = Create(PtrDeviceObject, PtrIrp);
		  break;

			case IRP_MJ_FILE_SYSTEM_CONTROL:
			Result = RiadenieFs(PtrDeviceObject, PtrIrp);
			break;

			case IRP_MJ_QUERY_VOLUME_INFORMATION:
			Result = QueryVolume(PtrDeviceObject, PtrIrp);
			break;

			case IRP_MJ_DEVICE_CONTROL:
			Result = RiadenieZariadenia(PtrDeviceObject, PtrIrp);
			break;

			case IRP_MJ_DIRECTORY_CONTROL:
			//Result = DirControl(PtrDeviceObject, PtrIrp);
			Result = QueryDirectory(PtrDeviceObject,PtrIrp);
			break;

			case IRP_MJ_CLOSE:
			Result = Close(PtrDeviceObject, PtrIrp);
			break;

			case IRP_MJ_CLEANUP:
			Result = Cleanup(PtrDeviceObject, PtrIrp);
			break;

			case IRP_MJ_READ:
			Result = Read(PtrDeviceObject, PtrIrp);
			break;

			case IRP_MJ_QUERY_INFORMATION:
			Result = QueryInformation(PtrDeviceObject,PtrIrp);
			break;

			default:
			PtrIrp->IoStatus.Status = STATUS_DRIVER_INTERNAL_ERROR;
			IoCompleteRequest(PtrIrp, IO_NO_INCREMENT);
			Result = STATUS_DRIVER_INTERNAL_ERROR;
		}
	}
	__except (FilterVynimky(GetExceptionCode()))
	{
		Result = ObsluhaVynimky(NULL,GetExceptionCode());
	}
	FsRtlExitFileSystem();
	return(Result);
} /* Rozhod */

/*********************************************************************************
*
* Procedra na ukonenie ovldaa.
*
*********************************************************************************/
VOID Koniec (PDRIVER_OBJECT  FfsDriver)

{
	UNICODE_STRING  DosDeviceName;

	__try
	{

		//IoDeleteDevice(GlobalData.PtrMainDeviceObject);
		RtlInitUnicodeString(&DosDeviceName, DOS_DEVICE_NAME);
		IoDeleteSymbolicLink(&DosDeviceName);
		ExDeleteResourceLite(&(GlobalData.MainResource));

#if !defined(TARGETING_Win2K)
		DoTraceMessage(FFS_INIT_DONE,"%s stopping",DRIVER_NAME);
                WPP_CLEANUP(FfsDriver);
#endif

	}

		__except (FilterVynimky(GetExceptionCode()))
	{
		ObsluhaVynimky(NULL,GetExceptionCode());
	}
} /* Koniec */

/******************************************************************************
*Na vstupe je blok ako fragment !!! zatial
******************************************************************************/
NTSTATUS
ReadFromVCB(PTR_T_VCB  PtrVCB,
            UINT64   FsBlock,
            UINT64   Offset,
            UINT32   Size,
            PVOID   PtrBuffer)
{
	NTSTATUS       Result = STATUS_SUCCESS;
  UINT64         Pomocna;
	Pomocna =fsbtodb(&(PtrVCB->SuperBlock),FsBlock)+PtrVCB->PartOfset;
	Pomocna = Pomocna * SECTOR_SIZE;
	Pomocna += Offset;

	Result = ReadPhysical(
	              PtrVCB->PtrParentVolume->PtrTargetDevice,
	              Pomocna,
	              Size,
	              PtrBuffer);
	return Result;
}
/*****************************************************************************
* Na zadane cislo inode vyhodi uz alokovanu FCB ak je inak null;
******************************************************************************/
PTR_T_FCB   SearchInode(PTR_T_INODE PtrInode)
{
	PLIST_ENTRY   PtrNextFCB =NULL;
	PTR_T_FCB     PtrResultFCB = NULL;
	PTR_T_FCB     PtrMyFCB =NULL;
	PTR_T_VCB     PtrVCB = NULL;

	PtrVCB = PtrInode->PtrVCB;
	if  (IsListEmpty (&(PtrVCB->HeadFCB))) return NULL;
	PtrNextFCB = (PtrVCB->HeadFCB.Flink);
	do
	{
		PtrMyFCB =  CONTAINING_RECORD(PtrNextFCB,T_FCB,NextFCB);
		if (PtrMyFCB->InodeNumber == PtrInode->InodeNum) return PtrMyFCB;
		PtrNextFCB = PtrNextFCB->Flink;
	}
	while (PtrNextFCB!=&(PtrVCB->HeadFCB));
	return NULL;
}
/******************************************************************************
* Overi ci Fscontext ukazuje dobre  na vstupe je fcb a ccb
* volajuci zamyka
******************************************************************************/
NTSTATUS CheckBoth(IN PVOID What)
{
	NTSTATUS  Result = STATUS_INVALID_PARAMETER;
	PTR_T_FCB PtrFCB;
	PTR_T_CCB PtrCCB;
	//PTR_T_VCB PtrVCB;


  __try
    {
      if (What==NULL)
      {
         DoTraceMessage(FFS_CHECKS,"Bad(NULL) CCB");
         __leave;
      }
      PtrCCB = What;
      if (PtrCCB->Magic != CCB_MAGIC)
      {
         DoTraceMessage(FFS_CHECKS,"Bad CCB magic");
         __leave;
      }
      PtrFCB = PtrCCB->PtrParentFCB;
      if (!(PtrFCB))
      {
         DoTraceMessage(FFS_CHECKS,"Bad(NULL) FCB");
         __leave;
      }
      if (PtrFCB->Magic != FCB_MAGIC)
      {
        DoTraceMessage(FFS_CHECKS,"Bad FCB magic");
        __leave;
      }
      // ked  je fcb a ccb dobre dalej by asi nemalo zmysel pozerat este dalej
      Result = STATUS_SUCCESS;
    }
  __except (((Result =GetExceptionCode())==STATUS_ACCESS_VIOLATION)  ?
             EXCEPTION_EXECUTE_HANDLER : EXCEPTION_CONTINUE_SEARCH)
  {
     DoTraceMessage(FFS_CHECKS,"Acces violoation in ccb or fcb");
     DoTraceMessage(FFS_EXCEPTION,"Catching Exception number %u",Result);
  }
  return Result;
}
/******************************************************************************
* Overi ci   na vstupe je device object  a irp
******************************************************************************/
//doplnit akokeby pokusny pristup k polozkam dat
NTSTATUS CheckDevIrp(IN  PDEVICE_OBJECT	 PtrDeviceObject,
	             IN  PIRP		 PtrIrp)
{
  NTSTATUS  Result = STATUS_INVALID_PARAMETER;

  __try
  {
    if (!(PtrIrp))
    {
       DoTraceMessage(FFS_CHECKS,"Bad(NULL) irp");
       __leave;
    }
    if (!(PtrDeviceObject))
    {
       DoTraceMessage(FFS_CHECKS,"Bad(NULL) DeviceObject");
       __leave;
    }
    Result = STATUS_SUCCESS;
  }
  __except (((Result =GetExceptionCode())==STATUS_ACCESS_VIOLATION)  ?
             EXCEPTION_EXECUTE_HANDLER : EXCEPTION_CONTINUE_SEARCH)
  {
     DoTraceMessage(FFS_CHECKS,"Acces violoation in irp or device object");
     DoTraceMessage(FFS_EXCEPTION,"Catching Exception number %u",Result);
  }
  return Result;
}